
class {{ data["name"] }}({{ data["implements"] }}):

	def __init__(self):
		raise Exception("cannot initialize like this. use the factory method")

{% for method in data["methods"] %}
	def {{ method }}:
		pass

{% endfor %}
