
ErrObjectNil = "object is nil"
ErrInvalidMethod = "method not allowed"
ErrObjectExists = "object already exists"
ErrNoSuchObject = "object does not exist"
ErrObjectIdentityMismatch = "object identity mismatch"
ErrInvalidFilter = "invalid filter key"
ErrInvalidPath = "invalid request path"
ErrInvalidRequest = "invalid request content"

DATETIME_FORMAT = "%Y-%m-%dT%H:%M:%S.%fZ"
